library(dplyr)
library(caret)
library(readr)
# install.packages("caret")


# load the data files
wine_red <- read_delim("data/winequality-red.csv", ";", escape_double = FALSE, trim_ws = TRUE)
wine_white <- read_delim("data/winequality-white.csv", ";", escape_double = FALSE, trim_ws = TRUE)

##
## STEP 1: exploring data
##

# data dimension
dim(wine_red)
dim(wine_white)

# data types
sapply(wine_red, class)
sapply(wine_white, class)

# let's add a column indicating red/white and join the 2 tables into 1
wine_red$type="RED"
wine_white$type="WHITE"

wine <- rbind(wine_red, wine_white)
head(wine)
View(wine)
sapply(wine, class)
wine$type <- as.factor(wine$type)
levels(wine$type)

# lets assess percentage of data for each wine type
percent <- prop.table(table(wine$type)) * 100
percent_type <- cbind(count=table(wine$type), percent=percent)
percent_type

# lets see the quality scores
hist(wine$quality)
# lets assign different qualities to 3 levels:
wine$quality_level="LO"
wine$quality_level[which(wine$quality>4)] = "MID"
wine$quality_level[which(wine$quality>6)] = "HI"
wine$quality_level <- as.factor(wine$quality_level)

percent <- prop.table(table(wine$quality_level)) * 100
percent_qual <- cbind(count=table(wine$quality_level), percent=percent)
percent_qual

# summarize all columns
summary(wine)

# lets use some graphics to better understand the values

par(mfcol=c(3,4), mar=c(1,1,1,1))
for (i in c(1:12)) {
  boxplot(wine[,i], main=names(wine)[i])
}


# look at svatterplots of all attributes
featurePlot(x=wine[,1:12], y=wine$type, plot="box")
featurePlot(x=wine[,1:4], y=wine$quality_level, plot="ellipse")

scales=list(x=list(relation="free"),y=list(relation="free"))
featurePlot(x=wine[,1:11], y=wine$quality_level, plot="density", scales=scales)


##
## STEP 2: peparining algorithms and method
##

control <- trainControl(method="cv", number=10)


# all caret compatible models available here:
# https://topepo.github.io/caret/available-models.html

# set seet for reproducibility
set.seed(2021)

# as we noticed earlier, the labels are not balanced. we need to balance them

percent_type
percent_qual


balanced_wine_type <- upSample(x=wine[,1:11],y=wine$type)
balanced_wine_qual <- upSample(x=wine[,1:11],y=wine$quality_level)

cbind(count=table(balanced_wine_type$Class), percent=prop.table(table(balanced_wine_type$Class)) * 100)
cbind(count=table(balanced_wine_qual$Class), percent=prop.table(table(balanced_wine_qual$Class)) * 100)


# redefine col names (solves a few issues in caret package)
colnames(balanced_wine_type) <- make.names(colnames(balanced_wine_type))
colnames(balanced_wine_qual) <- make.names(colnames(balanced_wine_qual))


# keep a portion of the data for testing the trained models

p_index_type <- createDataPartition(balanced_wine_type$Class, p = .8, list = FALSE)
p_index_qual <- createDataPartition(balanced_wine_qual$Class, p = .8, list = FALSE)
p_index_score <- createDataPartition(wine$quality, p = .8, list = FALSE)

test_wine_type <- balanced_wine_type[-p_index_type,]
test_wine_qual <- balanced_wine_qual[-p_index_qual,]
test_wine_score <- wine[-p_index_score,]

balanced_wine_type <- balanced_wine_type[p_index_type,]
balanced_wine_qual <- balanced_wine_qual[p_index_qual,]
wine <- wine[p_index_score,];

# define a few model algorithms

## CLASSIFICATION
metric="Accuracy"
# LDA 
fit.lda_type <- train(Class~., data=balanced_wine_type, method="lda", metric=metric, trControl=control)
fit.lda_qual <- train(Class~., data=balanced_wine_qual, method="lda", metric=metric, trControl=control)
# CART
fit.cart_type <- train(Class~., data=balanced_wine_type, method="rpart", metric=metric, trControl=control)
fit.cart_qual <- train(Class~., data=balanced_wine_qual, method="rpart", metric=metric, trControl=control)
# kNN
fit.knn_type <- train(Class~., data=balanced_wine_type, method="knn", metric=metric, trControl=control)
fit.knn_qual <- train(Class~., data=balanced_wine_qual, method="knn", metric=metric, trControl=control)
# SVM
fit.svm_type <- train(Class~., data=balanced_wine_type, method="svmRadial", metric=metric, trControl=control)
fit.svm_qual <- train(Class~., data=balanced_wine_qual, method="svmLinear", metric=metric, trControl=control)
# Random Forest
fit.rf_type <- train(Class~., data=balanced_wine_type, method="rf", metric=metric, trControl=control)
fit.rf_qual <- train(Class~., data=balanced_wine_qual, method="rf", metric=metric, trControl=control)

## REGRESSION
metric="RMSE"
# Linear Regression
fit.lm_score <- train(quality~., data=wine, method="lm", metric=metric, trControl=control)
# Penalized Linear Regression
fit.plm_score <- train(quality~., data=wine, method="penalized", metric=metric, trControl=control)
# L2 SVM with Linear kernel
fit.svm_score <- train(quality~., data=wine, method="svmLinear3", metric=metric, trControl=control)
# ElasticNet
fit.enet_score <- train(quality~., data=wine, method="enet", metric=metric, trControl=control)


##
## STEP 3: assessment
# assess all model performances

results_fit_type <- resamples(list(lda=fit.lda_type, cart=fit.cart_type, knn=fit.knn_type, svm=fit.svm_type, rf=fit.rf_type))
summary(results_fit_type)
bwplot(results_fit_type)

print(fit.rf_type)


results_fit_qual <- resamples(list(lda=fit.lda_qual, cart=fit.cart_qual, knn=fit.knn_qual, svm=fit.svm_qual, rf=fit.rf_qual))
summary(results_fit_qual)
bwplot(results_fit_qual)
print(fit.rf_qual)

results_fit_score <- resamples(list(lda=fit.lm_score, cart=fit.plm_score, knn=fit.svm_score, svm=fit.enet_score))
summary(results_fit_score)
bwplot(results_fit_score)
print(fit.svm_score)

# use ROC for a model
library(pROC)
probsTrain <- predict(fit.cart_type, test_wine_type, type = "prob")
rocCurve   <- roc(response = test_wine_type$Class,
                  predictor = probsTrain[, "RED"],
                  levels = rev(levels(test_wine_type$Class)))
plot(rocCurve, print.thres = "best")


# custom plot confusion matrix
conf_matrix <- function(df.true, df.pred, title = "", true.lab ="True Class", pred.lab ="Predicted Class",
                        high.col = 'red', low.col = 'white') {
  #convert input vector to factors, and ensure they have the same levels
  df.true <- as.factor(df.true)
  df.pred <- factor(df.pred, levels = levels(df.true))
  
  #generate confusion matrix, and confusion matrix as a pecentage of each true class (to be used for color) 
  df.cm <- table(True = df.true, Pred = df.pred)
  df.cm.col <- df.cm / rowSums(df.cm)
  
  #convert confusion matrices to tables, and binding them together
  df.table <- reshape2::melt(df.cm)
  df.table.col <- reshape2::melt(df.cm.col)
  df.table <- left_join(df.table, df.table.col, by =c("True", "Pred"))
  
  #calculate accuracy and class accuracy
  acc.vector <- c(diag(df.cm)) / c(rowSums(df.cm))
  class.acc <- data.frame(Pred = "Class Acc.", True = names(acc.vector), value = acc.vector)
  acc <- sum(diag(df.cm)) / sum(df.cm)
  
  #plot
  ggplot() +
    geom_tile(aes(x=Pred, y=True, fill=value.y),
              data=df.table, size=0.2, color=grey(0.5)) +
    geom_tile(aes(x=Pred, y=True),
              data=df.table[df.table$True==df.table$Pred, ], size=1, color="black", fill = 'transparent') +
    scale_x_discrete(position = "top",  limits = c(levels(df.table$Pred), "Class Acc.")) +
    scale_y_discrete(limits = rev(unique(levels(df.table$Pred)))) +
    labs(x=pred.lab, y=true.lab, fill=NULL,
         title= paste0(title, "\nAccuracy ", round(100*acc, 1), "%")) +
    geom_text(aes(x=Pred, y=True, label=value.x),
              data=df.table, size=4, colour="black") +
    geom_text(data = class.acc, aes(Pred, True, label = paste0(round(100*value), "%"))) +
    scale_fill_gradient(low=low.col, high=high.col, labels = scales::percent,
                        limits = c(0,1), breaks = c(0,0.5,1)) +
    guides(size=F) +
    theme_bw() +
    theme(panel.border = element_blank(), legend.position = "bottom",
          axis.text = element_text(color='black'), axis.ticks = element_blank(),
          panel.grid = element_blank(), axis.text.x.top = element_text(angle = 30, vjust = 0, hjust = 0)) +
    coord_fixed()
  
} 


conf_matrix(test_wine_qual$Class, predict(fit.lda_qual, test_wine_qual))
conf_matrix(test_wine_qual$Class, predict(fit.cart_qual, test_wine_qual))
conf_matrix(test_wine_qual$Class, predict(fit.knn_qual, test_wine_qual))
conf_matrix(test_wine_qual$Class, predict(fit.svm_qual, test_wine_qual))
conf_matrix(test_wine_qual$Class, predict(fit.rf_qual, test_wine_qual))


par(mfrow=c(1,2))
# generic residuals plot model
residuals = resid(fit.svm_score)
plot(wine$quality, residuals, 
     ylab="Residuals", xlab="Quality", 
     main="Wine quality prediction") 
abline(0, 0)     
qqnorm(residuals)
qqline(residuals)
